def match_chrom_format(chrom, keys):
    if chrom in keys:
        return chrom
    if "chr" in chrom:
        chrom2 = chrom.replace("chr", "")
    else:
        chrom2 = "chr{}".format(chrom)
        
    if chrom2 in keys:
        return chrom2
    return chrom


def render_to_file(doc, outf):
    """
    Renders the document to a file-like object.
    """
    for l in doc.render():
        outf.write(l + "\n")


import collections
import numpy

from genomeview.track import Track
from genomeview.axis import get_ticks

COLORS = ["blue", "red", "green", "black"]

class Series:
    def __init__(self, x, y, color=None, label=None):
        self.x = x
        self.y = y
        self.color = color
        self.label = label
        
class GraphTrack(Track):
    def __init__(self, name="graph", x=None, y=None):
        super().__init__(name)

        self.series = collections.OrderedDict()
        
        self.min_y = 0
        self.max_y = 0
        
        if x is not None:
            self.add_series(x, y)

        self.height = 100
        self.ymargin = 5
        
    def add_series(self, x, y, color=None, label=None):
        if label is None:
            label = "series_{}".format(len(self.series))
            
        assert label not in self.series

        x = numpy.asarray(x)
        y = numpy.asarray(y)

        if color is None:
            color = COLORS[len(self.series) % len(COLORS)]
            
        self.series[label] = Series(x, y, color, label)

        self.min_y = min(self.min_y, y[numpy.isfinite(y)].min())
        self.max_y = max(self.max_y, y[numpy.isfinite(y)].max())

    def ytopixels(self, yval):
        height = self.max_y - self.min_y
        return self.height - ((yval - self.min_y) / height * (self.height-2*self.ymargin) + self.ymargin)
        
    def render(self, renderer):
        for label, series in self.series.items():
            for i in range(len(series.x)-1):
                if any(numpy.isnan(series.x[i:i+2])) or any(numpy.isnan(series.y[i:i+2])):
                    continue
                x1 = self.scale.topixels(series.x[i])
                x2 = self.scale.topixels(series.x[i+1])
                y1 = self.ytopixels(series.y[i])
                y2 = self.ytopixels(series.y[i+1])
                                         
                yield from renderer.line(x1, y1, x2, y2, 
                    **{"stroke-width":1, "stroke":series.color, "stroke-linecap":"square"})

        # since the labels are drawn at the top of the ticks, let's make sure the top tick/label is 
        # more than 12 pixels from the top of the track so it doesn't get clipped
        # TODO: this ignores the margin, as of now
        axis_max_y = self.min_y + (self.max_y - self.min_y) * (1-7/self.height)
        ticks = get_ticks(self.min_y, axis_max_y, 4)

        yield from renderer.line(1, self.ytopixels(ticks[0][0]), 1, self.ytopixels(ticks[-1][0]), 
                                 **{"stroke-width":2, "stroke":"gray", "stroke-linecap":"square"})
        for tick, label in ticks:
            y = self.ytopixels(tick)
            yield from renderer.line(1, y, 10, y, 
                                     **{"stroke-width":2, "stroke":"gray", "stroke-linecap":"square"})
            yield from renderer.text(14, y, label, anchor="start", fill="gray")
            



import numpy

from libc.stdint cimport uint32_t, uint8_t, uint64_t, int64_t
from cpython cimport PyBytes_FromStringAndSize

from pysam.libcalignmentfile cimport AlignmentFile, AlignedSegment, IteratorRowAll, IteratorColumnRegion, IteratorRowRegion
# from pysam.calignmentfile cimport pysam_get_flag, pysam_bam_get_qname
from pysam.libcalignmentfile cimport BAM_FPROPER_PAIR, BAM_FPAIRED, BAM_FREVERSE, BAM_FMREVERSE, BAM_FREAD1, BAM_FUNMAP, BAM_FMUNMAP, BAM_FDUP, pysam_bam_get_qname
from pysam.libcalignedsegment cimport PileupColumn, pysam_bam_get_seq
from pysam.libchtslib cimport bam1_t, bam_pileup1_t

from genomeview.utilities import match_chrom_format


cdef char* bam_nt16_rev_table = "=ACMGRSVTWYHKDBN"

cdef inline object get_seq_base(bam1_t *src, uint32_t k):
    cdef uint8_t* p
    cdef char* s

    if not src.core.l_qseq:
        return None

    seq = PyBytes_FromStringAndSize(NULL, 1)
    s   = <char*>seq
    p   = pysam_bam_get_seq(src)

    s[0] = bam_nt16_rev_table[p[k//2] >> 4 * (1 - k%2) & 0xf]

    return seq


class MismatchCounts(object):
    """
    keeps track of how many of each nucleotide (or insertion/deletion) are present at each position
    -> used for the quick consensus mode
    """
    types_to_id = {"A":0, "C":1, "G":2, "T":3, "DEL":5}

    def __init__(self, chrom, start, end):
        self.chrom = chrom
        self.start = start
        self.end = end

        length = end - start
        self.counts = numpy.zeros([6, length])#, dtype="uint8")
        self.insertions = numpy.zeros(length)#, dtype="uint8")

    def tally_reads(self, bam):
        cdef:
            IteratorColumnRegion pileup
            PileupColumn column
            bam_pileup1_t ** plp
            bam_pileup1_t * pileupRead
            bam1_t *read
            int c, n

        chrom = match_chrom_format(self.chrom, bam.references)
	
        # for pileupcolumn in bam.pileup(self.chrom, self.start, self.end, truncate=True):
        pileup = bam.pileup(chrom, self.start, self.end, truncate=True)
        for column in pileup:
            # for pileupread in pileupcolumn.pileups:
            n = column.n
            plp = column.plp

            for i in range(n):
                pileupRead = &(plp[0][i])
                read = pileupRead.b

                if pileupRead.is_refskip:
                    continue
                elif pileupRead.is_del:
                    self.add_count(column.pos, "DEL")
                else:
                    # nuc =  pileupRead.alignment.query_sequence[pileupread.query_position]
                    nuc = get_seq_base(read, pileupRead.qpos)
                    if nuc != "N":
                        self.add_count(column.pos, nuc.decode())
                if pileupRead.indel > 0:
                    self.add_count(column.pos, "INS")

    def add_count(self, position, type_):
        position -= self.start
        if type_ == "INS":
            self.insertions[position] += 1
        else:
            if type_ not in self.types_to_id:
                return
            row = self.types_to_id[type_]
            self.counts[row,position] += 1

    # def counts(self, position):
    #     position -= self.start
    #     return self.counts[position,:]

    def query(self, type_, start, end=None):
        if start < self.start or start >= self.end:
            return False

        start -= self.start
        if end is None:
            end = start
        else:
            end -= self.start

        total = self.counts[:,start:(end+1)].sum(axis=0)
        total = total.astype(float)

        if (type_ == "INS"):
            insertions = self.insertions[start:(end+1)]
            if (insertions.sum() / total.sum()) > 0.2:
               return True

        else:        
            row = self.types_to_id[type_]
            this_type = self.counts[row,start:(end+1)]

            if type_ == "DEL":
                if  ((this_type / total) > 0.3).any():
                    return True
            elif ((this_type / total) > 0.2).any():
                return True

        return False



from genomeview.track import Track

class Interval:
    def __init__(self, id_, chrom, start, end, strand="+", label=None):
        self.chrom = chrom
        self.start = start
        self.end = end
        self.id = id_
        self.label = label
        
        if isinstance(strand, bool):
            strand = {True:"+", False:"-"}[strand]
        self.strand = strand

    def overlaps(self, other, ignore_strand=True):
        if self.chrom != other.chrom:
            return False
        if other.start > self.end or self.start > other.end:
            return False
        if not ignore_strand and self.strand != other.strand:
            return False
        return True

    def __repr__(self):
        return "{}:{:,}-{:,}{}".format(self.chrom, self.start, self.end, self.strand)
        
def color_by_strand(interval):
    # brightness = 0.2 + (cur_reads[0].mapq/40.0*0.8)
    color = "#E89E9D"
    if interval.strand == "-":
        color = "#8C8FCE"
    return color
    
class IntervalTrack(Track):
    def __init__(self, name, intervals):
        super().__init__(name)
        self.rows = []
        self.intervals_to_rows = {}
        
        self.row_height = 8
        self.margin_x = 15
        self.margin_y = 2
        self.label_distance = 3
        
        self.intervals = intervals

        self.color_fn = color_by_strand

    def layout_interval(self, interval):
        row = 0
        interval_start = self.scale.topixels(interval.start)
        for row, row_end in enumerate(self.rows):
            if interval_start > row_end:
                break
        else:
            self.rows.append(None)
            row = len(self.rows) - 1
        
        new_end = self.scale.topixels(interval.end) + self.margin_x
        if interval.label is not None:
            new_end += len(interval.label) * self.row_height * 0.75
        self.rows[row] = new_end

        assert not interval.id in self.intervals_to_rows
        self.intervals_to_rows[interval.id] = row
        

    def layout(self, scale):
        super().layout(scale)

        self.rows = []
        self.intervals_to_rows = {}

        for interval in self.intervals:
            self.layout_interval(interval)
            
        self.height = (len(self.rows)+1) * (self.row_height+self.margin_y)
    
    def draw_interval(self, renderer, interval):
        start = self.scale.topixels(interval.start)
        end = self.scale.topixels(interval.end)
        
        row = self.intervals_to_rows[interval.id]
        top = row*(self.row_height+self.margin_y)
        
        color = self.color_fn(interval)
        temp_label = interval.label
        if interval.label is None:
            temp_label = "{}_{}".format(interval.id, 1 if interval.read.is_read1 else 2)
        yield from renderer.rect(start, top, end-start, self.row_height, fill=color, **{"stroke":"none", "id":temp_label})
        if interval.label is not None:
            yield from renderer.text(end+self.label_distance, top+self.row_height-2, interval.label, anchor="start")
        
    def render(self, renderer):
        for interval in self.intervals:
            yield from self.draw_interval(renderer, interval)
            
        yield from self.render_label(renderer)



import genomeview

def test_basic(reference_path):
    source = genomeview.FastaGenomeSource(reference_path)

    doc = genomeview.Document(900)
    
    view = genomeview.GenomeView("locus 1", "chr4", 96549060, 96549060+1000, "+", source)
    doc.add_view(view)

    bam_track_hg002 = genomeview.SingleEndBAMTrack("HG002", "quick_consensus_test.bam")
    view.add_track(bam_track_hg002)

    axis_track = genomeview.Axis("axis")
    view.add_track(axis_track)

    with open("output.svg", "w") as out_svg_file:
        genomeview.render_to_file(doc, out_svg_file)

    import os
    assert os.path.exists("output.svg")


"""
Genome Sources are used to describe reference genomes, typically either on disk (eg
:class:`genomeview.genomesource.FastaGenomeSource`) or in memory 
(:class:`genomeview.genomesource.GenomeSource`).
"""

import collections

from genomeview.utilities import match_chrom_format

comp = str.maketrans('ATCGNatcgn','TAGCNtagcn')
    
def reverse_comp(st):
    """ Returns the reverse complement of a DNA sequence; non ACGT bases will be ignored. """
    return str(st)[::-1].translate(comp)



class GenomeSource(object):
    """
    An in-memory genome source.

    Attributes:
        names_to_contigs: a `collections.OrderedDict` or similar ordered mapping specifying 
            the contig/chromosome names as keys and their sequences as values.
    """
    def __init__(self, names_to_contigs, aligner_type="bwa"):
        self.names_to_contigs = collections.OrderedDict(names_to_contigs)
        self._bwa = None
        self._ssw = None
        self._blacklist = None

        self.aligner_type = aligner_type

    def get_seq(self, chrom, start, end, strand):
        chrom = match_chrom_format(chrom, self.keys())
        seq = self.names_to_contigs[chrom][start:end+1]
        if strand == "-":
            seq = reverse_comp(seq)
        return seq

    def keys(self):
        return list(self.names_to_contigs.keys())



class FastaGenomeSource(GenomeSource):
    """ 
    A genome source based on a Fasta file. 

    This is essentially a pickle-able wrapper for pyfaidx.Fasta
    """
    def __init__(self, path):
        self.path = path
        self._fasta = None
        
    def get_seq(self, chrom, start, end, strand):
        chrom = match_chrom_format(chrom, self.keys())
        
        seq = self.fasta[chrom][start:end+1]
        if strand == "-":
            seq = reverse_comp(seq)
        return seq

    def keys(self):
        return list(self.fasta.keys())

    @property
    def fasta(self):
        if self._fasta is None:
            import pyfaidx
            self._fasta = pyfaidx.Fasta(self.path, as_raw=True)
        return self._fasta

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_fasta"] = None
        return state



__version__ = "0.9.4.2"

from genomeview.genomeview import *
from genomeview.genomesource import *

try:
    from genomeview._quickconsensus import *
except ImportError:
    import logging
    logging.warn("Unable to load cythonized quickconsensus module; "
                 "drawing reads using quick consensus mode will be "
                 "substantially slower")
    from genomeview.quickconsensus import *

from genomeview.axis import *
from genomeview.track import *
from genomeview.bamtrack import *
from genomeview.graphtrack import *
from genomeview.intervaltrack import *

from genomeview.utilities import render_to_file
